@@ -65,6 +65,12 @@ class Normalise:
         # Contains "action" and "verbose"
         errors_found = False
 
+        def is_value_array(key, value) -> bool:
+            if type(rules[j][key]) != list:
+                print(f'{key} expects an array but got "{value}" maybe try ["{value}"]')
+                return False
+            return True
+
         # Gets called when row doesn't match rule
         def handle_mismatch(i, value, rows):
             nonlocal errors_found
@@ -86,6 +92,7 @@ class Normalise:
                 # Which rows to drop later onin one go
                 drop_array.append(i)
 
+            # Feedback management to user
             verbose = global_verbose
             if "verbose" in rules[j]:
                 verbose = rules[j]["verbose"]
@@ -107,6 +114,27 @@ class Normalise:
                 nonlocal global_mapping
                 nonlocal dataframe
 
+                # Change all column names
+                if "normalise_columns" in rules[j]:
+                    if is_value_array(key = "normalise_columns", value = rules[j]["normalise_columns"]) is False:
+                        return
+
+                    for option in rules[j]["normalise_columns"]:
+                        if option == "lowercase":
+                            dataframe.columns = dataframe.columns.str.lower()
+                        elif option == "uppercase":
+                            dataframe.columns = dataframe.columns.str.upper()
+                        elif option == ("capitalize", "capitalise"):
+                            dataframe.columns = dataframe.columns.str.capitalize()
+                        elif option == "remove_special":
+                            dataframe.columns = dataframe.columns.str.replace(r'[^A-Za-z0-9 ]+', '', regex=True)
+                        elif option == "remove_whitespace":
+                            dataframe.columns = dataframe.columns.str.replace(r'\s+', '', regex=True)
+                        elif option == "spaces_to_underscore":
+                            dataframe.columns = dataframe.columns.str.replace(r'\s+', '_', regex=True)
+                        elif option == "spaces_to_hyphen":
+                            dataframe.columns = dataframe.columns.str.replace(r'\s+', '-', regex=True)
+
                 if "check_coverage" in rules[j]:
                     global_check_coverage = rules[j]["check_coverage"]
 
@@ -193,7 +221,51 @@ class Normalise:
                 print("Column " + rules[j]["column"] + " not found")
                 continue
 
-            if "mapping" in rules[j] or global_mapping != False:
+            if "normalise" in rules[j]:
+                if is_value_array(key = "normalise", value = rules[j]["normalise"]) is False:
+                    continue
+
+                for option in rules[j]["normalise"]:
+                    
+                    if option == "lowercase":
+                        column_values = column_values.str.lower()
+                    elif option == "uppercase":
+                        column_values = column_values.str.upper()
+                    elif option in ("capitalize", "capitalise"):
+                        column_values = column_values.str.capitalize()
+                    elif option == "remove_special":
+                        column_values = column_values.str.replace(r'[^a-zA-Z0-9]', '', regex=True)
+                    elif option == "remove_whitespace":
+                        column_values = column_values.str.replace(r'\s+', '', regex=True)
+                    elif option == "spaces_to_underscore":
+                        column_values = column_values.str.replace(r'\s+', '_', regex=True)
+                    elif option == "spaces_to_hyphen":
+                        column_values = column_values.str.replace(r'\s+', '-', regex=True)
+
+            # Same as global rule but now only for one column
+            if "normalise_columns" in rules[j]:
+                if is_value_array(key = "normalise_columns", value = rules[j]["normalise_columns"]) is False:
+                    continue
+
+                for option in rules[j]["normalise_columns"]:
+                    print(rules[j]["normalise_columns"])
+                    if option == "lowercase":
+                        dataframe.rename(columns={rules[j]["column"]: rules[j]["column"].lower()}, inplace=True)
+                    elif option == "uppercase":
+                        dataframe.rename(columns={rules[j]["column"]: rules[j]["column"].upper()}, inplace=True)
+                    elif option in ("capitalize", "capitalise"):
+                        dataframe.rename(columns={rules[j]["column"]: rules[j]["column"].capitalize()}, inplace=True)
+                    elif option == "remove_special":
+                        dataframe.rename(columns={rules[j]["column"]: re.sub(r'[^a-zA-Z0-9]', '', rules[j]["column"])}, inplace=True)
+                    elif option == "remove_whitespace":
+                        dataframe.rename(columns={rules[j]["column"]: re.sub(r'\s+', '', rules[j]["column"])}, inplace=True)
+                    elif option == "spaces_to_underscore":
+                        dataframe.rename(columns={rules[j]["column"]: re.sub(r'\s+', '_', rules[j]["column"])}, inplace=True)
+                    elif option == "spaces_to_hyphen":
+                        dataframe.rename(columns={rules[j]["column"]: re.sub(r'\s+', '-', rules[j]["column"])}, inplace=True)
+                continue
+
+            if "mapping" in rules[j] or global_mapping is not False:
                 mapping = global_mapping
                 if "mapping" in rules[j]:
                     mapping = rules[j]["mapping"]
@@ -241,12 +313,14 @@ class Normalise:
                     except ValueError:
                         handle_mismatch(i, value, column_values)
             
+            # Make the data inside the column Panda timestamps
             if "timestamp" in rules[j]:
                 try:
                     dataframe[rules[j]["column"]] = pd.to_datetime(dataframe[rules[j]["column"]], utc=True).dt.strftime(rules[j]["timestamp"])
                 except TypeError:
                     print("Timestamping failed: " + str(rules[j]["column"]))
 
+            # Check if a column lays between to dates
             if "range_time" in rules[j]:
                 left = pd.to_datetime(rules[j]["range_time"][0], utc=True)
                 right = pd.to_datetime(rules[j]["range_time"][1], utc=True)
@@ -259,6 +333,7 @@ class Normalise:
                     except ValueError:
                         handle_mismatch(i, value, column_values)
 
+            # Check if a column matches a specific format
             if "type" in rules[j]:
                     
                 if rules[j]["type"] == "percentage":
@@ -314,6 +389,7 @@ class Normalise:
                             handle_mismatch(i, value, column_values)
 
                 if rules[j]["type"] == "postal_code":
+                    # Remove special characters and make uppercase
                     column_values.astype(str).str.replace(
                         r"[^a-zA-Z0-9]+", "", regex=True).str.upper()
 
@@ -335,6 +411,7 @@ class Normalise:
                     for i, value in column_values.items():
                         value = list(value)
                         for idx, character in enumerate(value):
+                            # Remove all non-alphanumeric except ' - and whitespaces
                             if character not in ("'", "-", " ") and character.isalpha() is False:
                                 handle_mismatch(i, value, column_values)
                                 break
@@ -343,12 +420,13 @@ class Normalise:
 
                             if character in ("'", "-"):
                                 try:
-                                    # Make next character uppercase
+                                    # Make next character uppercase after ' or -
                                     value[idx + 1] = value[idx + 1].upper()
                                 except IndexError:
                                     continue
 
                             column_values[i] = "".join(value)
+
             # Loop again because error in sample
             if (reset_coverage and errors_found and len(column_values) < len(dataframe[rules[j]["column"]])):
                 j -= 1
@@ -369,6 +447,7 @@ class Normalise:
                     array_name = 'dropped: ' if array == drop_array else 'made None: ' if array == none_array else 'made NaN: '
                     print(f"Column {rules[j]['column']}, {array_name}{len(array)} values")
 
+            # Fill the NaN values with something else
             try:
                 if "fillna" in rules[j]:
                     column_values = column_values.fillna(rules[j]["fillna"])
